import asyncio
import time
from typing import List, Dict, Any

import aiohttp
from more_itertools import chunked
from tqdm import tqdm


class TentacleSession(aiohttp.ClientSession):
    """ TentacleSession is a wrapper around the aiohttp.ClientSession, where it introduces the retry and rate functionality
     missing in the default aiohttp.ClientSession.

        Args:
            sleep (float): The time the client will sleep after each request. \n
      	    retries (int): The number of retries for a successful request. \n
     	    retry_sleep (float): The time service sleeps between nonsuccessful request calls. Defaults to 1.0.

        Returns:
            TentacleSession(aiohttp.ClientSession)
    """
    sleep: float
    retries: int
    retry_sleep: float = 1.0

    def __init__(self, sleep: float, retries=3, retry_sleep=1.0, **kwargs):
        self.retries = retries
        self.sleep = sleep
        self.retry_sleep = retry_sleep
        super().__init__(raise_for_status=True, **kwargs)

    def __retry__(self, func, **kwargs) -> Any:
        attempts = 0
        error = Exception()
        while attempts < self.retries:
            try:
                start_time = time.time()
                resp = func(**kwargs)
                response_time = round(time.time() - start_time, 1)
                if response_time < self.sleep:
                    time.sleep(self.sleep - response_time)
                return resp
            except Exception as error:
                attempts += 1
                error = error
                time.sleep(self.retry_sleep)

        raise error

    def get(self, **kwargs) -> Any:
        return self.__retry__(func=super().get, **kwargs)

    def patch(self, **kwargs) -> Any:
        return self.__retry__(func=super().patch, **kwargs)

    def post(self, **kwargs) -> Any:
        return self.__retry__(func=super().post, **kwargs)

    def put(self, **kwargs) -> Any:
        return self.__retry__(func=super().put, **kwargs)

    def request(self, **kwargs) -> Any:
        return self.__retry__(func=super().request, **kwargs)


class OctopusApi:
    """ Initiates the Octopus client.
        Args:
            rate (Optional[float]): The rate limits of the endpoint; default to no limit. \n
 	        resolution (Optional[str]): The time resolution of the rate (sec, minute), defaults to None.
	        concurrency (int): Maximum concurrency on the given endpoint, defaults to 30; if rate is provided,
	        then concurrency will be set to 1.

        Returns:
            OctopusApi
    """
    rate_sec: float = None
    concurrency: int
    retries: int

    def __init__(self, rate: int = None, resolution: str = None, concurrency: int = 30,
                 retries: int = 3):

        if rate or resolution:
            if resolution.lower() not in ["minute", "sec"]:
                raise ValueError("Incorrect value of resolution, expecting minute or sec!")
            if not rate:
                raise ValueError("Can not set resolution of rate without rate")
            self.rate_sec = rate / (60 if resolution.lower() == "minute" else 1)

        self.concurrency = concurrency if not rate else 1
        self.retries = retries

    def execute(self, requests_list: List[Dict[str, Any]], func: callable) -> List[Any]:
        """ Execute the requests given the functions instruction.

            Empower asyncio libraries for performing parallel executions of the user-defined function.
            Given a list of requests, the result is an out of order list of what the user-defined function returns.

            Args:
                requests_list (List[Dict[str, Any]): The list of requests in a dictionary format, e.g.
                [{"url": "http://example.com", "params": {...}, "body": {...}}..]
                func (callable): The user-defined function to execute, this function takes in the following arguments.
                    Args:
                        session (TentacleSession): The Octopus wrapper around the aiohttp.ClientSession.
                        request (Dict): The request within the requests_list above.

            Returns:
                List(func->return)
        """

        async def __tentacles__(rate: float, retries: int, concurrency: int, requests_list: List[Dict[str, Any]],
                                func: callable) -> List[Any]:
            sleep: float = 0
            if rate:
                sleep = round(1 / rate, 2)
            conn = aiohttp.TCPConnector(limit_per_host=concurrency)
            async with TentacleSession(sleep=sleep, retries=retries, connector=conn) as session:
                return await asyncio.gather(
                    *[func(session, request) for request in requests_list])

        batches = chunked(requests_list, self.concurrency)
        results = list()
        for batch in tqdm(list(batches)):
            r = asyncio.run(__tentacles__(self.rate_sec, self.retries, self.concurrency, batch, func))
            results.extend(r)

        return results



# -*- coding: utf-8 -*-
from setuptools import setup

modules = \
['octopus_api']
install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'asyncio>=3.4.3,<4.0.0',
 'more-itertools>=8.11.0,<9.0.0',
 'tqdm>=4.62.3,<5.0.0']

entry_points = \
{'console_scripts': ['octopus = octopus:octopus']}

setup_kwargs = {
    'name': 'octopus-api',
    'version': '1.0.0',
    'description': "Octopus-api is a python library for performing client-based optimized concurrent requests and limit rates set by the endpoint contract. Octopus-api is simple; it combines the asyncio and aiohttp library's functionality and makes sure the requests follow the constraints set by the user.",
    'long_description': '# octopus-api\n![octopus_icon](https://github.com/FilipByren/octopus-api/blob/main/image.png?raw=true)\n## About\nOctopus-api is a python library for performing client-based optimized concurrent requests and limit rates set by the endpoint contract.\n\nOctopus-api is simple; it combines the [asyncio](https://docs.python.org/3/library/asyncio.html) and [aiohttp](https://docs.aiohttp.org/en/stable/) library\'s functionality and makes sure the requests follows the constraints set by the contract.\n\n## Installation\n`pip install octopus-api`\n\n## PyPi\nhttps://pypi.org/project/octopus-api/\n\n\n## Get started\nTo start Octopus, you first initiate the client, setting your constraints. \n```python\nclient = OctopusApi(rate=30, resolution="minute", retries=10)\nclient = OctopusApi(rate=5, resolution="sec", retries=3)\nclient = OctopusApi(concurrency=100, retries=5)\n```\nAfter that, you will specify what you want to perform on the endpoint response. This is done within a user-defined function.\n```python\nasync def patch_data(session: TentacleSession, request: Dict):\n    async with session.patch(url=request["url"], data=requests["data"], params=request["params"]) as response:\n        body = await response.json()\n        return body["id"]\n```\n\nAs Octopus `TentacleSession` uses [aiohttp](https://docs.aiohttp.org/en/stable/) under the hood, the resulting  way to write \n**POST**, **GET**, **PUT** and **PATCH** for aiohttp will be the same for octopus. The only difference is the added functionality of \nretries and rate limits (if set).\n\nFinally, you finish everything up with the `execute` call for the octopus client, where you provide the list of requests dicts your defined and the user function. The execute call will then return the waited list of returned values.\n\n```python\nresult: List = client.execute(requests_list=[\n    {\n        "url": "http://localhost:3000",\n        "data": {"id": "a", "first_name": "filip"},\n        "params": {"id": "a"}\n    },\n    {\n        "url": "http://localhost:3000",\n        "data": {"id": "b", "first_name": "morris"},\n        "params": {"id": "b"} \n    }\n    ] , func=patch_data)\n```\n\n\n### Examples\n\nOptimize the request based concurrency constraints:\n```python\nfrom octopus_api import TentacleSession, OctopusApi\nfrom typing import Dict, List\n\nif __name__ == \'__main__\':\n    async def get_text(session: TentacleSession, request: Dict):\n        async with session.get(url=request["url"], params=request["params"]) as response:\n            body = await response.text()\n            return body\n\n\n    client = OctopusApi(concurrency=100)\n    result: List = client.execute(requests_list=[{\n        "url": "http://google.com",\n        "params": {}}] * 100, func=get_text)\n    print(result)\n\n```\nOptimize the request based on rate limit constraints:\n```python\nfrom octopus_api import TentacleSession, OctopusApi\nfrom typing import Dict, List\n\nif __name__ == \'__main__\':\n    async def get_ethereum_id(session: TentacleSession, request: Dict):\n        async with session.get(url=request["url"], params=request["params"]) as response:\n            body = await response.json()\n            return body["id"]\n\n    client = OctopusApi(rate=30, resolution="minute")\n    result: List = client.execute(requests_list=[{\n        "url": "http://api.coingecko.com/api/v3/coins/ethereum?tickers=false&localization=false&market_data=false",\n        "params": {}}] * 100, func=get_ethereum_id)\n    print(result)\n\n```\n\n## Limitations\n1. Returned result from the user defined function comes in out of order.\n2. When setting rate-limiting, the api will not utilize the concurrency functionality and only run a single \noperation in sequence.\n',
    'author': 'Filip Byrén',
    'author_email': 'filip.j.byren@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/FilipByren/octopus-api',
    'py_modules': modules,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
